# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import abc
import atexit
import json
import tempfile
from hashlib import md5
from pathlib import Path
from typing import Any, Iterator, Union
import requests


class IndexHandler(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __getitem__(self, index) -> Any:
        ...

    @abc.abstractmethod
    def __len__(self) -> int:
        ...

    @abc.abstractmethod
    def __iter__(self) -> Iterator:
        ...

    @abc.abstractmethod
    def find(self) -> Union[list, None]:
        ...

    @abc.abstractmethod
    def update(self) -> bool:
        ...

    @property
    @abc.abstractmethod
    def outdated(self) -> bool:
        ...


class JsonFileIndexHandler(IndexHandler):
    """Local file handler."""
    def __init__(self, path: str) -> None:
        """With the given path, process the data and return the results."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else str(path)
        self._path = Path(path)
        if not self._path.exists():
            message = 'index file not found'
            raise FileNotFoundError(message)
        with open(self._path) as json_file:
            self._data = json.load(json_file)

    def __getitem__(self, index) -> any:
        """Get an item from the _data list read from the json file."""
        return self._data.__getitem__(index)

    def __len__(self) -> int:
        """Return the length of the _data list returned from the json file."""
        return self._data.__len__()

    def __iter__(self) -> Iterator:
        """Return the list of packages contained within the index."""
        return iter(self._data)

    def find(self, value: str) -> Union[list, None]:
        """Find packages matching a specified value."""
        return [s for s in self._data if value.lower() in s.lower()]

    def update(self) -> bool:
        """As the file is referenced directly, there is no need to update the
        index.
        """
        return True

    @property
    def outdated(self) -> bool:
        """As the file is referenced directly, there is no need to check the
        age.
        """
        return False


class JsonUrlIndexHandler(IndexHandler):
    """Use a json file located on a remote server."""
    def __init__(self, index: str, cache_dir: str = None) -> None:
        """With the given path, process the data and return the results."""
        self._index = index
        if not cache_dir:
            cache_dir = tempfile.TemporaryDirectory()
            atexit.register(cache_dir.cleanup)
        self._cache = Path(cache_dir)
        self._cache.mkdir(0x777, True, True)
        self._cached_index = Path(self._cache, 'index.json')
        self._cached_index.touch(exist_ok=True)
        self._updated = False
        self.update()
        with open(self._cached_index) as json_file:
            self._data = json.load(json_file)

    def __getitem__(self, index) -> Any:
        """Get an item from the _data list read from the json file."""
        return self._data.__getitem__(index)

    def __len__(self) -> int:
        """Return the length of the _data list returned from the json file."""
        return self._data.__len__()

    def __iter__(self) -> Iterator:
        """Return the list of packages contained within the index."""
        return iter(self._data)

    def find(self, value: str) -> Union[list, None]:
        """Find packages matching a specified value."""
        return [s for s in self._data if value.lower() in s.lower()]

    def update(self) -> bool:
        """Using the requests module, download the latest index."""
        self._updated = False
        headers = {'If-None-Match': self._digest}
        req = requests.get(self._index, headers=headers, allow_redirects=True)
        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            with open(self._cached_index, 'wb') as cached_index:
                cached_index.write(req.content)
            self._updated = True

    @property
    def outdated(self) -> bool:
        """Execute a head request using the requests library to check that the
        ETag matches.
        """
        headers = {'If-None-Match': self._digest}
        req = requests.head(self._index, headers=headers, allow_redirects=True)
        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            return True
        elif req.status_code == 304:
            return False

    @property
    def _digest(self) -> str:
        """Return the md5 digest of the currently cached index file."""
        digest = md5()  # noqa: S303
        with open(self._cached_index, "rb") as cached_index:
            for block in iter(lambda: cached_index.read(65536), b""):
                digest.update(block)
        return digest.hexdigest()



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from json import JSONDecodeError
import pytest
from dismantle.index import IndexHandler, JsonFileIndexHandler


def test_notfound(datadir):
    with pytest.raises(FileNotFoundError):
        JsonFileIndexHandler(datadir.join('index_notfound.json'))


def test_blank(datadir):
    with pytest.raises(JSONDecodeError):
        JsonFileIndexHandler(datadir.join('index_blank.json'))


def test_empty(datadir):
    index = JsonFileIndexHandler(datadir.join('index_empty.json'))
    assert len(index) == 0


def test_broken(datadir):
    with pytest.raises(JSONDecodeError):
        JsonFileIndexHandler(datadir.join('index_broken.json'))


def test_populated(datadir):
    index = JsonFileIndexHandler(datadir.join('index_populated.json'))
    assert isinstance(index, IndexHandler) is True


def test_rfc8089(datadir):
    path = 'file://' + str(datadir.join('index_populated.json'))
    index = JsonFileIndexHandler(path)
    assert isinstance(index, IndexHandler) is True


def test_outdated(datadir):
    index = JsonFileIndexHandler(datadir.join('index_populated.json'))
    assert index.outdated is False


def test_update(datadir):
    index = JsonFileIndexHandler(datadir.join('index_populated.json'))
    assert index.update() is True


def test_length(datadir):
    index = JsonFileIndexHandler(datadir.join('index_populated.json'))
    assert len(index) == 6
    assert len(index.find('@scope-one/package-one')) == 1
    assert len(index.find('package-one')) == 3
    assert len(index.find('package-two')) == 2
    assert len(index.find('package-three')) == 1
    assert len(index.find('@scope-one')) == 3
    assert len(index.find('@scope-two')) == 2
    assert len(index.find('@scope-three')) == 1


def test_populated_package_exists(datadir):
    index = JsonFileIndexHandler(datadir.join('index_populated.json'))
    package = index["@scope-one/package-one"]
    assert package["name"] == "@scope-one/package-one"
    assert package["version"] == "0.1.0"
    assert package["path"] == "@scope-one/package-one"


def test_populated_package_nonexistant(datadir):
    index = JsonFileIndexHandler(datadir.join('index_populated.json'))
    with pytest.raises(KeyError):
        index["@scope-four/package-one"]



#      888 ,e,                                     d8   888
#  e88 888  "   dP"Y 888 888 8e   ,"Y88b 888 8e   d88   888  ,e e,
# d888 888 888 C88b  888 888 88b "8" 888 888 88b d88888 888 d88 88b
# Y888 888 888  Y88D 888 888 888 ,ee 888 888 888  888   888 888   ,
#  "88 888 888 d,dP  888 888 888 "88 888 888 888  888   888  "YeeP"
# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""Dismantle is a Python package / plugin / extension manager."""
import logging
from dismantle.__version__ import __version__  # noqa: F401


logging.getLogger(__name__).addHandler(logging.NullHandler())



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""Dismantle package manager system with support for extensions and plugins."""

# IDEAS: --- plugins ----------------------------
# IDEAS: Function hook to provide plugin (process content before / after)
# IDEAS: List plugins for a function

_plugins = {}

from .iplugin import IPlugin
from .plugin import plugin
from .plugins import Plugins
from .register import register


__all__ = [
    'IPlugin',
    'register',
    'plugin',
    'Plugins',
    '_plugins'
]



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import abc
import shutil
import tarfile
import zipfile
from pathlib import Path


class PackageFormat(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def grasps(path: any) -> bool:
        """Return a boolean value describing whether the Format can be
        processed.
        """
        ...

    @staticmethod
    @abc.abstractmethod
    def verify(signature: str) -> bool:
        """Verify a packages hash with the provided signature."""
        ...

    @staticmethod
    @abc.abstractmethod
    def extract(src: any, dest: str) -> bool:
        """Verify a packages hash with the provided signature."""
        ...


class DirectoryPackageFormat(PackageFormat):

    @staticmethod
    def grasps(path: any) -> bool:
        """Check if a directory on the local filesystem has been provided."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        return Path(str(path)).is_dir()

    @staticmethod
    def extract(src: any, dest: str) -> bool:
        """Use the formatter to process any movement related actions."""
        src = str(src)[7:] if str(src)[:7] == 'file://' else src
        dest = str(dest)[7:] if str(dest)[:7] == 'file://' else dest
        if not DirectoryPackageFormat.grasps(src):
            message = 'formatter only supports directories'
            raise ValueError(message)
        if dest != src:
            try:
                shutil.copytree(src, dest)
            except FileExistsError:
                raise FileExistsError('destination already exists')


class ZipPackageFormat(PackageFormat):

    @staticmethod
    def grasps(path: any) -> bool:
        """Check if a directory on the local filesystem has been provided."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        zip_path = Path(path)
        if zip_path.suffix != '.zip':
            return False
        return True

    @staticmethod
    def extract(src: str, dest: str) -> bool:
        """Extract the zipfile to the cache location if a copy doesn't already
        exist."""
        src = str(src)[7:] if str(src)[:7] == 'file://' else src
        dest = str(dest)[7:] if str(dest)[:7] == 'file://' else dest
        src = Path(src)
        if not ZipPackageFormat.grasps(src):
            message = 'formatter only supports zip files'
            raise ValueError(message)
        if not src.is_file() or not zipfile.is_zipfile(src):
            message = 'invalid zip file'
            raise ValueError(message)
        dest_path = Path(dest)
        if dest_path.exists():
            message = 'destination already exists'
            raise FileExistsError(message)
        with zipfile.ZipFile(src, 'r') as zip_ref:
            zip_ref.extractall(dest_path)


class TarPackageFormat(PackageFormat):

    @staticmethod
    def grasps(path: any) -> bool:
        """Check if a directory on the local filesystem has been provided."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        tar_path = Path(path)
        suffixes = ''.join(tar_path.suffixes)
        if suffixes != '.tar':
            return False
        return True

    @staticmethod
    def extract(src: str, dest: str) -> bool:
        """Extract the tarfile to the cache location if a copy doesn't already
        exist."""
        src = str(src)[7:] if str(src)[:7] == 'file://' else src
        dest = str(dest)[7:] if str(dest)[:7] == 'file://' else dest
        src = Path(src)
        if not TarPackageFormat.grasps(src):
            message = 'formatter only supports tar files'
            raise ValueError(message)
        if not src.is_file() or not tarfile.is_tarfile(src):
            message = 'invalid tar file'
            raise ValueError(message)
        dest_path = Path(dest)
        if dest_path.exists():
            message = 'destination already exists'
            raise FileExistsError(message)
        with tarfile.open(src, 'r') as tar_ref:
            tar_ref.extractall(dest_path)


class TgzPackageFormat(PackageFormat):

    @staticmethod
    def grasps(path: any) -> bool:
        """Check if a directory on the local filesystem has been provided."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        tgz = Path(path)
        suffixes = ''.join(tgz.suffixes)
        if suffixes not in ['.tgz', '.tar.gz']:
            return False
        return True

    @staticmethod
    def extract(src: str, dest: str) -> bool:
        """Extract the tarfile to the cache location if a copy doesn't already
        exist."""
        src = str(src)[7:] if str(src)[:7] == 'file://' else src
        dest = str(dest)[7:] if str(dest)[:7] == 'file://' else dest
        src = Path(src)
        if not TgzPackageFormat.grasps(src):
            message = 'formatter only supports tar.gz files'
            raise ValueError(message)
        if not src.is_file() or not tarfile.is_tarfile(src):
            message = 'invalid tgz file'
            raise ValueError(message)
        dest_path = Path(dest)
        if dest_path.exists():
            message = 'destination already exists'
            raise FileExistsError(message)
        with tarfile.open(src, 'r') as tgz_ref:
            tgz_ref.extractall(dest_path)



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Check packages."""
import argparse
import sys
from . import __version__


def main(args=None):
    """Main command line entry point."""
    args = args if args else sys.argv[1:]
    parser = argparse.ArgumentParser(
        'dismantle',
        description='Dismantle package management'
    )
    app = '%(prog)s version ' + __version__
    parser.add_argument('-v', '--version', action='version', version=app)
    args = parser.parse_args(args)



#      888 ,e,                                     d8   888
#  e88 888  "   dP"Y 888 888 8e   ,"Y88b 888 8e   d88   888  ,e e,
# d888 888 888 C88b  888 888 88b "8" 888 888 88b d88888 888 d88 88b
# Y888 888 888  Y88D 888 888 888 ,ee 888 888 888  888   888 888   ,
#  "88 888 888 d,dP  888 888 888 "88 888 888 888  888   888  "YeeP"

# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import shutil
from pathlib import Path
import pytest


@pytest.fixture(scope='function')
def datadir(tmpdir, request):
    """Fixture responsible for searching a folder with the same name of test
    module and, if available, moving all contents to a temporary directory so
    tests can use them freely.
    """
    filename = Path(request.module.__file__)
    test_dir = Path(filename.parent, 'data', filename.stem)

    if test_dir.is_dir():
        shutil.rmtree(str(tmpdir), ignore_errors=True)
        shutil.copytree(test_dir, str(tmpdir))

    return tmpdir


@pytest.fixture(scope="session")
def httpserver_listen_address():
    """Use port 9090 for testing."""
    return ("127.0.0.1", 9090)



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import os
from pathlib import Path
import pytest
from dismantle.package import DirectoryPackageFormat, PackageFormat


def test_inherits() -> None:
    assert issubclass(DirectoryPackageFormat, PackageFormat) is True


def test_grasp_exists(datadir: Path) -> None:
    src = datadir.join('directory_src')
    assert DirectoryPackageFormat.grasps(src) is True


def test_grasp_non_existant(datadir: Path) -> None:
    src = datadir.join('directory_non_existant')
    assert DirectoryPackageFormat.grasps(src) is False


def test_grasp_not_supported(datadir: Path) -> None:
    src = datadir.join('package.zip')
    assert DirectoryPackageFormat.grasps(src) is False


def test_extract_not_supported(datadir: Path) -> None:
    src = datadir.join('package.zip')
    dest = datadir.join(f'{src}_output')
    message = 'formatter only supports directories'
    with pytest.raises(ValueError, match=message):
        DirectoryPackageFormat.extract(src, dest)


def test_extract_non_existant(datadir: Path) -> None:
    src = datadir.join('directory_non_existant')
    dest = datadir.join(f'{src}_output')
    message = 'formatter only supports directories'
    with pytest.raises(ValueError, match=message):
        DirectoryPackageFormat.extract(src, dest)


def test_extract_already_exists(datadir: Path) -> None:
    src = datadir.join('directory_src')
    dest = datadir.join('directory_exists')
    message = 'destination already exists'
    with pytest.raises(FileExistsError, match=message):
        DirectoryPackageFormat.extract(src, dest)


def test_extract_create(datadir: Path) -> None:
    src = datadir.join('directory_src')
    dest = datadir.join('directory_created')
    DirectoryPackageFormat.extract(src, dest)
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""Extensions locator."""
import importlib.util
import os
import sys
from contextlib import suppress
from pathlib import Path
from . import IExtension


class Extensions:
    """Search through the installed packages and find extensions."""
    def __init__(self, types, packages, prefix) -> None:
        """Search through all provided extensions and register them."""
        self._packages = packages
        self._extensions = {}
        self._directory = 'extensions'
        self._prefix = prefix
        self._imports = {}
        self._exclude = (['__pycache__', '.DS_Store'])
        self._include = ('.py', '.pyc')
        # check that the types are a subclass of IExtension
        if not all([issubclass(i, IExtension) for i in types]):
            raise ValueError('all exntesion types must extend IExtension')
        self._types = types
        for category in self._types:
            self._extensions[category._category] = {}
        self._find()
        self._register()

    def _find(self) -> None:
        """Search through the packages and find all extensions. """
        for package in self._packages.values():
            # set the package prefix
            _prefix = package.name
            # check if the package has an init file
            with suppress(KeyError):
                root, paths, _ = next(os.walk(package._path))
            # check if we have an extensions directory
            if self._directory in paths:
                ext_files = os.path.join(root, self._directory)
                for root, dirs, files in os.walk(ext_files, topdown=True):
                    dirs[:] = [d for d in dirs if d not in self._exclude]
                    files[:] = [f for f in files if f.endswith(self._include)]
                    for name in files:
                        # get the path from the beginning of the module
                        fd = os.path.splitext(name)[0]
                        ext_length = len(ext_files)
                        full_path = os.path.join(root, fd)
                        path = full_path[ext_length:].replace(os.sep, '.')
                        prefix = _prefix + '.extension' + path
                        self._imports[prefix] = self._load(full_path, prefix)
                        self._imports[prefix].prefix = prefix

    def _load(self, path, prefix):
        """Python 3.5 and up. (I hate you package system)."""
        ending = '/__init__.py' if os.path.isdir(path) else '.py'
        path = Path(f'{path}{ending}')
        spec = importlib.util.spec_from_file_location(prefix, path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[prefix] = module
        try:
            spec.loader.exec_module(module)
        except Exception as e:  # NOQA: E722
            raise ModuleNotFoundError(f'error in {prefix} ({e})')
        return module

    def _register(self) -> None:
        """Register extensions."""
        for i in self._imports.values():
            for cls, cls_name in ((getattr(i, name), name) for name in dir(i)):
                if isinstance(cls, type):
                    for subclass in self._types:
                        if issubclass(cls, subclass) and cls is not subclass:
                            name = '.'.join([i.prefix, cls_name])
                            self._extensions[subclass._category][name] = cls

    def category(self, category) -> list:
        return self._extensions[category]

    @property
    def types(self) -> list:
        return list(self._extensions.keys())

    @property
    def extensions(self) -> dict:
        return self._extensions

    @property
    def imports(self) -> dict:
        return self._imports



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import sys
from pathlib import Path
from dismantle.extension import Extensions
from dismantle.index import JsonFileIndexHandler
from dismantle.package import LocalPackageHandler


def test_success(datadir: Path) -> None:
    from tests.ColorExtension import ColorExtension
    from tests.GreetingExtension import GreetingExtension
    ext_types = [ColorExtension, GreetingExtension]
    index_src = datadir.join('index.json')
    index = JsonFileIndexHandler(index_src)
    packages = {}
    for pkg_meta in index:
        meta = index[pkg_meta]
        path = datadir.join(meta['path'])
        package = LocalPackageHandler(meta['name'], path)
        package._meta = {**package._meta, **meta}
        package.install()
        packages[package.name] = package
    extensions = Extensions(ext_types, packages, 'd_')
    assert extensions.types == ['color', 'greeting']
    assert list(extensions.category('color').keys()) == [
        '@scope-one/package-one.extension.green.GreenColorExtension',
        '@scope-one/package-two.extension.red.RedColorExtension',
        '@scope-one/package-three.extension.blue.BlueColorExtension',
    ]
    assert list(extensions.extensions.keys()) == ['color', 'greeting']
    assert list(extensions.imports.keys()) == [
        '@scope-one/package-one.extension.hello',
        '@scope-one/package-one.extension.green',
        '@scope-one/package-two.extension.afternoon',
        '@scope-one/package-two.extension.red',
        '@scope-one/package-three.extension.goodbye',
        '@scope-one/package-three.extension.blue'
    ]


def test_in_sys_modules(datadir: Path) -> None:
    from tests.ColorExtension import ColorExtension
    ext_types = [ColorExtension]
    index_src = datadir.join('index_in_sys_modules.json')
    index = JsonFileIndexHandler(index_src)
    packages = {}
    for pkg_meta in index:
        meta = index[pkg_meta]
        path = datadir.join(meta['path'])
        package = LocalPackageHandler(meta['name'], path)
        package._meta = {**package._meta, **meta}
        package.install()
        packages[package.name] = package
    extensions = Extensions(ext_types, packages, 'd_')
    assert extensions.types == ['color']
    assert list(extensions.category('color').keys()) == [
        '@scope-five/package-one.extension.carbon.CarbonColorExtension'
    ]
    assert list(extensions.extensions.keys()) == ['color']
    assert list(extensions.imports.keys()) == [
        '@scope-five/package-one.extension.carbon'
    ]
    assert '@scope-five/package-one.extension.carbon' in sys.modules



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from hashlib import md5
from json import JSONDecodeError
from shutil import copy2
import pytest
from pytest_httpserver import HTTPServer
from pytest_httpserver.httpserver import HandlerType
from requests import ConnectionError
from dismantle.index import IndexHandler, JsonUrlIndexHandler


def test_invalid_server(httpserver: HTTPServer, tmpdir):
    with pytest.raises(ConnectionError):
        JsonUrlIndexHandler('http://invalid.server/notfound.json', tmpdir)
    httpserver.check_assertions()


def test_notfound(httpserver: HTTPServer, tmpdir):
    httpserver.no_handler_status_code = 404
    params = {"uri": "/invalid.json", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data("")
    with pytest.raises(FileNotFoundError):
        JsonUrlIndexHandler(httpserver.url_for('notfound.json'), tmpdir)


def test_blank(httpserver: HTTPServer, tmpdir):
    params = {"uri": "/blank.json", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data("")
    with pytest.raises(JSONDecodeError):
        JsonUrlIndexHandler(httpserver.url_for('blank.json'), tmpdir)
    httpserver.check_assertions()


def test_empty(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_empty.json')) as json_file:
        data = json_file.read()
    params = {"uri": "/empty.json", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data(data)
    index = JsonUrlIndexHandler(httpserver.url_for('empty.json'), datadir)
    httpserver.check_assertions()
    assert len(index) == 0


def test_broken(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_broken.json')) as json_file:
        data = json_file.read()
    params = {"uri": "/broken.json", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data(data)
    with pytest.raises(JSONDecodeError):
        JsonUrlIndexHandler(httpserver.url_for('broken.json'), datadir)
    httpserver.check_assertions()


def test_populated(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_populated.json')) as json_file:
        data = json_file.read()
    params = {"uri": "/populated.json", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data(data)
    index = JsonUrlIndexHandler(httpserver.url_for('populated.json'), datadir)
    assert isinstance(index, IndexHandler) is True
    httpserver.check_assertions()


def test_latest(httpserver: HTTPServer, datadir):
    copy2(datadir.join('index_populated.json'), datadir.join('index.json'))
    file_digest = md5()  # noqa: S303
    with open(datadir.join('index_populated.json'), "rb") as cached_index:
        for block in iter(lambda: cached_index.read(65536), b""):
            file_digest.update(block)
    digest = file_digest.hexdigest()
    headers = {"If-None-Match": digest}
    params = {
        "uri": "/latest.json",
        "headers": headers,
        'handler_type': HandlerType.ONESHOT
    }
    httpserver.expect_request(**params).respond_with_data("", status=304)
    index = JsonUrlIndexHandler(httpserver.url_for('latest.json'), datadir)
    assert isinstance(index, IndexHandler) is True
    assert index._updated is False


def test_current(httpserver: HTTPServer, datadir):
    copy2(datadir.join('index_populated.json'), datadir.join('index.json'))
    file_digest = md5()  # noqa: S303
    with open(datadir.join('index_populated.json'), "rb") as cached_index:
        for block in iter(lambda: cached_index.read(65536), b""):
            file_digest.update(block)
    digest = file_digest.hexdigest()
    headers = {"If-None-Match": digest}
    params = {
        "uri": "/current.json",
        "headers": headers,
        'handler_type': HandlerType.PERMANENT
    }
    httpserver.expect_request(**params).respond_with_data("", status=304)
    index = JsonUrlIndexHandler(httpserver.url_for('current.json'), datadir)
    assert isinstance(index, IndexHandler) is True
    assert index.outdated is False


def test_outdated(httpserver: HTTPServer, datadir):
    copy2(datadir.join('index_empty.json'), datadir.join('index.json'))
    file_digest = md5()  # noqa: S303
    with open(datadir.join('index_empty.json'), "rb") as cached_index:
        for block in iter(lambda: cached_index.read(65536), b""):
            file_digest.update(block)
    digest = file_digest.hexdigest()
    headers = {"If-None-Match": digest}
    params = {
        "uri": "/outdated.json",
        "headers": headers,
        "method": "GET",
        'handler_type': HandlerType.ONESHOT
    }
    httpserver.expect_request(**params).respond_with_data("", status=304)
    index = JsonUrlIndexHandler(httpserver.url_for('outdated.json'), datadir)
    assert isinstance(index, IndexHandler) is True
    assert index._updated is False
    params['method'] = 'HEAD'
    with open(datadir.join('index_populated.json')) as json_file:
        data = json_file.read()
    httpserver.expect_request(**params).respond_with_data(data, status=200)
    assert index.outdated is True


def test_create(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_populated.json')) as json_file:
        data = json_file.read()
    httpserver.expect_request(
        "/create.json"
    ).respond_with_data(data, status=200)
    index = JsonUrlIndexHandler(httpserver.url_for('create.json'), datadir)
    assert isinstance(index, IndexHandler) is True
    assert index._updated is True


def test_update(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_populated.json')) as json_file:
        data = json_file.read()
    httpserver.expect_request(
        "/update.json"
    ).respond_with_data(data, status=200)
    index = JsonUrlIndexHandler(httpserver.url_for('update.json'), datadir)
    assert isinstance(index, IndexHandler) is True
    assert index._updated is True


def test_populated_index_length(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_populated.json')) as json_file:
        data = json_file.read()
    httpserver.expect_request("/populated.json").respond_with_data(data)
    index = JsonUrlIndexHandler(httpserver.url_for('populated.json'), datadir)
    httpserver.check_assertions()
    assert len(index) == 6
    assert len(index.find('@scope-one/package-one')) == 1
    assert len(index.find('package-one')) == 3
    assert len(index.find('package-two')) == 2
    assert len(index.find('package-three')) == 1
    assert len(index.find('@scope-one')) == 3
    assert len(index.find('@scope-two')) == 2
    assert len(index.find('@scope-three')) == 1


def test_populated_pkg_exists(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_populated.json')) as json_file:
        data = json_file.read()
    params = {"uri": "/exists.json", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data(data)
    index = JsonUrlIndexHandler(httpserver.url_for('exists.json'), datadir)
    httpserver.check_assertions()
    package = index["@scope-one/package-one"]
    assert package["name"] == "@scope-one/package-one"
    assert package["version"] == "0.1.0"
    assert package["path"] == "@scope-one/package-one"


def test_populated_pkg_nonexistant(httpserver: HTTPServer, datadir):
    with open(datadir.join('index_populated.json')) as json_file:
        data = json_file.read()
    params = {"uri": "/nonexist.json", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data(data)
    index = JsonUrlIndexHandler(httpserver.url_for('nonexist.json'), datadir)
    with pytest.raises(KeyError):
        index["@scope-four/package-one"]
    httpserver.check_assertions()



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""Plugins locator."""

import logging
import os


try:
    from importlib.abc import Loader as imp
except ImportError:
    import imp  # NOQA: F401

from . import IPlugin, _plugins


# enable logging
log = logging.getLogger(__name__)


class Plugins:
    """Search through the installed packages and find plugins."""

    def __init__(self, types, packages, prefix):
        """ """
        self._packages = packages
        self._plugins = _plugins
        self._directory = 'plugins'
        self._prefix = prefix
        self._imports = {}
        self._exclude = set(['__pycache__'])

        # check that the types are a subclass of IPlugin
        if all([issubclass(i, IPlugin) for i in types]):
            self._types = types
        else:
            raise ValueError('all plugin types must extend IPlugin')
        self._find()

    def _find(self):
        """search through the packages and find all plugins."""
        for x in self._packages.enabled.values():
            _prefix = x['name']
            # check if the package has an init file
            try:
                root, paths, filenames = next(os.walk(x['directory']))
            except KeyError:
                continue
            # check if we have an extensions directory
            if self._directory in paths:
                plg_files = os.path.join(root, self._directory)
                for root, dirs, files in os.walk(plg_files, topdown=True):
                    dirs[:] = [d for d in dirs if d not in self._exclude]

                    for name in files:
                        # get the path from the beginning of the module
                        fd = os.path.splitext(name)[0]
                        full_path = os.path.join(root, fd)
                        v = full_path[len(plg_files):].replace(os.sep, '.')
                        prefix = _prefix + '.plugin' + v
                        self._imports[prefix] = self._import(full_path, prefix)

    def _import(self, path, prefix):
        """ Import a file or module into our imports list. """

        # use imp to correctly load the plugin as a module
        if os.path.isdir(path):
            spec = ("py", "r", imp.PKG_DIRECTORY)
            module = imp.load_module(prefix, None, path, spec)
        else:
            with open(path + ".py", "r") as plugin_file:
                spec = ("py", "r", imp.PY_SOURCE)
                module = imp.load_module(
                    prefix,
                    plugin_file,
                    path + ".py",
                    spec
                )
        return module

    @property
    def plugins(self):
        return self._plugins

    @property
    def imports(self):
        return self._imports



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import os
from pathlib import Path
import pytest
from pytest_httpserver import HTTPServer
from pytest_httpserver.httpserver import HandlerType
from requests import ConnectionError
from dismantle.package import (
    DirectoryPackageFormat,
    HttpPackageHandler,
    PackageHandler,
    TarPackageFormat,
    TgzPackageFormat,
    ZipPackageFormat
)


def test_inherits() -> None:
    assert issubclass(HttpPackageHandler, PackageHandler) is True


def test_grasp_http_support() -> None:
    src = 'http://google.com/package.zip'
    assert HttpPackageHandler.grasps(src) is True


def test_grasp_https_support() -> None:
    src = 'https://google.com/package.zip'
    assert HttpPackageHandler.grasps(src) is True


def test_grasp_invalid_ftp() -> None:
    src = 'ftp://google.com/package.zip'
    assert HttpPackageHandler.grasps(src) is False


def test_grasp_invalid_file(datadir: Path) -> None:
    src = datadir.join('package.zip')
    assert HttpPackageHandler.grasps(src) is False


def test_grasp_invalid_path(datadir: Path) -> None:
    src = datadir.join('@scope-one/package-one')
    assert HttpPackageHandler.grasps(src) is False


def test_invalid_server(httpserver: HTTPServer) -> None:
    name = '@scope-one/package-one'
    src = 'http://invalid.server.com/notfound.zip'
    package = HttpPackageHandler(name, src)
    with pytest.raises(ConnectionError):
        package.install(src, '0.0.1')
    httpserver.check_assertions()


def test_notfound(httpserver: HTTPServer, datadir) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('notfound.zip')
    dest = datadir.join('package-nonexistant')
    httpserver.no_handler_status_code = 404
    params = {"uri": "/invalid.zip", 'handler_type': HandlerType.ONESHOT}
    httpserver.expect_request(**params).respond_with_data("")
    package = HttpPackageHandler(name, src)
    with pytest.raises(FileNotFoundError):
        package.install(dest)


def test_install_dir_exists(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    dest = 'directory_exists'
    message = 'destination already exists'
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.installed is False
    with pytest.raises(FileExistsError, match=message):
        package.install(datadir.join(dest), '0.0.1')
    assert package.installed is False


def test_install_create(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    dest = datadir.join('package-create')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(package._cache)


def test_meta_value_nonexistant(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    dest = datadir.join('package-create')
    message = 'nonexistant is an invalid attribute'
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.installed is False
    assert package.install(dest, '0.0.1') is True
    with pytest.raises(AttributeError, match=message):
        assert package.nonexistant


def test_name(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    dest = datadir.join('package-create')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.installed is False
    assert package.install(dest, '0.0.1') is True
    assert package.name == name
    assert package.name != '@scope-two/package-one'


def test_name_invalid(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-two'
    src = httpserver.url_for('/package.zip')
    dest = datadir.join('package-create')
    message = 'meta name does not match provided package name'
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.installed is False
    with pytest.raises(ValueError, match=message):
        package.install(dest, '0.0.1')
    assert package.installed is False


def test_name_missing(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/missing_name.zip')
    dest = datadir.join('package-create')
    message = 'meta file missing name value'
    with open(datadir.join('missing_name.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/missing_name.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    with pytest.raises(ValueError, match=message):
        package.install(dest, '0.0.1')


def test_version(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    dest = datadir.join('package-version')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.install(dest, '0.0.1') is True
    assert package.version == '0.0.1'
    assert package.version != '1.0.1'


def test_version_missing(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/missing_version.zip')
    dest = datadir.join('package-create')
    message = 'meta file missing version value'
    with open(datadir.join('missing_version.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/missing_version.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    with pytest.raises(ValueError, match=message):
        package.install(dest, '0.0.1')


def test_verification_none(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.verify() is True


def test_verification_value(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    message = 'the http package handler does not support verification'
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    with pytest.raises(ValueError, match=message):
        package.verify('a0aea27ca371ef0e715c594300e22ef9')


def test_uninstall(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    dest = datadir.join('package-uninstall')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src)
    assert package.installed is False
    package.install(dest, version='0.0.1')
    assert package.installed is True
    assert package.uninstall() is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_zip_format(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    dest = datadir.join('package-zip')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    package = HttpPackageHandler(name, src, [ZipPackageFormat])
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_tar_format(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.tar')
    dest = datadir.join('package-tar')
    with open(datadir.join('package.tar'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.tar").respond_with_data(data)
    package = HttpPackageHandler(name, src, [TarPackageFormat])
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_tgz_format(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.tgz')
    dest = datadir.join('package-tgz')
    with open(datadir.join('package.tgz'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.tgz").respond_with_data(data)
    package = HttpPackageHandler(name, src, [TgzPackageFormat])
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_multi_format(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.tgz')
    dest = datadir.join('package-tgz')
    formats = [
        DirectoryPackageFormat,
        TgzPackageFormat,
        ZipPackageFormat,
        TarPackageFormat
    ]
    with open(datadir.join('package.tgz'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.tgz").respond_with_data(data)
    package = HttpPackageHandler(name, src, formats)
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_no_format(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    message = 'a valid source is required'
    with pytest.raises(FileNotFoundError, match=message):
        HttpPackageHandler(name, src, [])


def test_missing_format(httpserver: HTTPServer, datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = httpserver.url_for('/package.zip')
    with open(datadir.join('package.zip'), 'rb') as pkg_file:
        data = pkg_file.read()
    httpserver.expect_request("/package.zip").respond_with_data(data)
    message = 'a valid source is required'
    with pytest.raises(FileNotFoundError, match=message):
        HttpPackageHandler(name, src, [TarPackageFormat])



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
class IExtension:
    """Extension interface to recognise what needs to be processed."""

    _name: str = "default"
    _category: str = "default"

    def __init__(self) -> None:
        self._active: bool = False

    def activeate(self) -> bool:
        self._active = True
        return self._active

    def deactivate(self) -> bool:
        self._active = False
        return not self._active

    @property
    def name(self) -> str:
        return self._name

    @property
    def category(self) -> str:
        return self._category

    @property
    def active(self) -> bool:
        return self._active



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import os
from pathlib import Path
import pytest
from dismantle.package import (
    LocalPackageHandler,
    PackageHandler,
    DirectoryPackageFormat,
    TarPackageFormat,
    TgzPackageFormat,
    ZipPackageFormat
)


def test_inherits() -> None:
    assert issubclass(LocalPackageHandler, PackageHandler) is True


def test_grasp_exists(datadir: Path) -> None:
    src = datadir.join('package.zip')
    assert LocalPackageHandler.grasps(src) is True


def test_grasp_non_existant(datadir: Path) -> None:
    src = datadir.join('non_existant.zip')
    assert LocalPackageHandler.grasps(src) is False


def test_grasp_not_supported(datadir: Path) -> None:
    src = datadir.join('directory_src')
    assert LocalPackageHandler.grasps(src) is False


def test_grasp_not_http_url(datadir: Path) -> None:
    src = 'https://google.com/package.zip'
    assert LocalPackageHandler.grasps(src) is False


def test_grasp_file_url(datadir: Path) -> None:
    src = f'file://{datadir.join("package.zip")}'
    assert LocalPackageHandler.grasps(src) is True


def test_grasp_not_supported_format(datadir: Path) -> None:
    src = datadir.join('invalid.file')
    assert LocalPackageHandler.grasps(src) is False


def test_src_notfound(datadir: Path) -> None:
    name = 'package/not-found'
    src = datadir.join(name)
    message = 'unable to process source format'
    with pytest.raises(FileNotFoundError, match=message):
        LocalPackageHandler(name, src)


def test_subclass_correct(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    package = LocalPackageHandler(name, src)
    assert isinstance(package, PackageHandler) is True


def test_install_no_destination(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    package = LocalPackageHandler(name, src)
    assert package.installed is False
    assert package.install(version='0.0.1') is True
    assert package.installed is True


def test_install_directory_exists(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    dest = datadir.join('directory_exists')
    message = 'destination already exists'
    package = LocalPackageHandler(name, src)
    assert package.installed is False
    with pytest.raises(FileExistsError, match=message):
        package.install(dest, '0.0.1')
    assert package.installed is False


def test_install_directory_create(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    dest = datadir.join('package-nonexistant')
    package = LocalPackageHandler(name, src)
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(src)


def test_meta_value_nonexistant(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    package = LocalPackageHandler(name, src)
    message = 'nonexistant is an invalid attribute'
    assert package.install(src, '0.0.1') is True
    with pytest.raises(AttributeError, match=message):
        assert package.nonexistant


def test_name(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    package = LocalPackageHandler(name, src)
    assert package.install(src, '0.0.1') is True
    assert package.name == name
    assert package.name != '@scope-two/package-one'


def test_name_invalid(datadir: Path) -> None:
    name = '@scope-one/package-two'
    wrong_name = '@scope-one/package-one'
    src = datadir.join(wrong_name)
    message = 'meta name does not match provided package name'
    package = LocalPackageHandler(name, src)
    with pytest.raises(ValueError, match=message):
        package.install(src, '0.0.1')


def test_name_missing(datadir: Path) -> None:
    name = '@scope-one/package-three'
    src = datadir.join(name)
    message = 'meta file missing name value'
    package = LocalPackageHandler(name, src)
    with pytest.raises(ValueError, match=message):
        package.install(src, '0.0.1')


def test_version(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    package = LocalPackageHandler(name, src)
    assert package.install(src, '0.0.1') is True
    assert package.version == '0.0.1'
    assert package.version != '1.0.1'


def test_version_missing(datadir: Path) -> None:
    name = '@scope-one/package-four'
    src = datadir.join(name)
    message = 'meta file missing version value'
    package = LocalPackageHandler(name, src)
    with pytest.raises(ValueError, match=message):
        package.install(src, '0.0.1')


def test_verification_none(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    package = LocalPackageHandler(name, src)
    assert package.verify() is True


def test_verification_value(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    message = 'the local package handler does not support verification'
    package = LocalPackageHandler(name, src)
    with pytest.raises(ValueError, match=message):
        package.verify('a0aea27ca371ef0e715c594300e22ef9')


def test_uninstall(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    dest = datadir.join('test_uninstall')
    package = LocalPackageHandler(name, src)
    assert package.installed is False
    package.install(dest, version='0.0.1')
    assert package.installed is True
    assert os.path.exists(src)
    assert package.uninstall() is True
    assert os.path.exists(src) is True
    assert package.installed is False


def test_uninstall_with_dest(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    dest = datadir.join('package-nonexistant')
    package = LocalPackageHandler(name, src)
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest)
    assert package.uninstall() is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_directory_format(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join(name)
    dest = datadir.join('package-nonexistant')
    package = LocalPackageHandler(name, src)
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest)
    assert package.uninstall() is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_zip_format(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join('package.zip')
    dest = datadir.join('package-nonexistant')
    package = LocalPackageHandler(name, src, [ZipPackageFormat])
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_tar_format(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join('package.tar')
    dest = datadir.join('package-nonexistant')
    package = LocalPackageHandler(name, src, [TarPackageFormat])
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_tgz_format(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join('package.tgz')
    dest = datadir.join('package-nonexistant')
    package = LocalPackageHandler(name, src, [TgzPackageFormat])
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_multi_format(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join('package.tgz')
    dest = datadir.join('package-nonexistant')
    formats = [
        DirectoryPackageFormat,
        TgzPackageFormat,
        ZipPackageFormat,
        TarPackageFormat
    ]
    package = LocalPackageHandler(name, src, formats)
    assert package.installed is False
    package.install(dest, '0.0.1')
    assert package.installed is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is True
    assert os.path.exists(dest / 'package.json') is True
    assert package.uninstall() is True
    assert os.path.exists(src) is True
    assert os.path.exists(dest) is False
    assert package.installed is False


def test_no_format(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join('package.zip')
    message = 'unable to process source format'
    with pytest.raises(FileNotFoundError, match=message):
        LocalPackageHandler(name, src, [])


def test_missing_format(datadir: Path) -> None:
    name = '@scope-one/package-one'
    src = datadir.join('package.zip')
    message = 'unable to process source format'
    with pytest.raises(FileNotFoundError, match=message):
        LocalPackageHandler(name, src, [TarPackageFormat])



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import abc
import atexit
import json
import shutil
import tempfile
from hashlib import md5
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import List, Optional
from urllib.parse import urlparse
import requests
from ._formats import DirectoryPackageFormat, PackageFormat, ZipPackageFormat


Formats = Optional[List[PackageFormat]]


class PackageHandler(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __getattr__(self, name) -> object:
        ...

    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def installed(self) -> bool:
        ...

    @staticmethod
    @abc.abstractmethod
    def grasps(path: any) -> bool:
        """Return a boolean value describing weither the package source can be
        processed.
        """
        ...

    @abc.abstractmethod
    def install(self, path: str, version: str = None) -> bool:
        """Install a specific package version."""
        ...

    @abc.abstractmethod
    def uninstall(self) -> bool:
        """Uninstall a package."""
        ...

    @abc.abstractmethod
    def verify(self, signature: str) -> bool:
        """Verify a packages hash with the provided signature."""
        ...


class LocalPackageHandler(PackageHandler):
    """Directory package structure."""
    def __init__(self, name: str, src: str, formats: Optional[Formats] = None):
        """initialise the package."""
        self._meta = {}
        self._meta['name'] = name
        self._path = None
        self._installed = False
        self._src = str(src)[7:] if str(src)[:7] == 'file://' else src
        if formats is None:
            formats = [DirectoryPackageFormat]
        for current_format in formats:
            if current_format.grasps(self._src):
                self._format = current_format
                break
        else:
            message = 'unable to process source format'
            raise FileNotFoundError(message)

    @property
    def name(self) -> str:
        """Return the name of the package from the meta data."""
        return self.meta['name']

    @property
    def path(self) -> str:
        """Return the path the package is installed into."""
        return self._path

    def __getattr__(self, name):
        """Return an attribute from the meta data if the data doesnt exist."""
        if name not in self._meta:
            message = f'{name} is an invalid attribute'
            raise AttributeError(message)
        return self._meta[name]

    @property
    def installed(self) -> bool:
        """Return the current installation state."""
        return self._installed

    @staticmethod
    def grasps(path: any) -> bool:
        """Check if a directory on the local filesystem has been provided."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        return Path(str(path)).exists()

    def install(self, path: str = None, version: str = None) -> bool:
        """The local package handler does not install the package. No version
        control exists for the directory package type.
        """
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        self._path = path if path else self._src
        self._format.extract(self._src, self._path)
        self._meta = {**self._meta, **self._load_metadata(self._path)}
        self._installed = True
        return True

    def uninstall(self) -> bool:
        """Uninstall the package."""
        if self._path != self._src:
            self._remove_files(self._path)
        self._path = None
        self._installed = False
        return True

    def verify(self, digest: str = None) -> bool:
        """Verify the package hasn't been tampered with."""
        if digest is None:
            return True
        message = 'the local package handler does not support verification'
        raise ValueError(message)

    def _load_metadata(self, path: Path):
        """Load the package.json file into memory."""
        path = Path(str(path)[7:] if str(path)[:7] == 'file://' else path)
        try:
            with open(path / 'package.json') as package:
                meta = json.load(package)
                if 'name' not in meta:
                    message = 'meta file missing name value'
                    raise ValueError(message)
                if self._meta['name'] != meta['name']:
                    message = 'meta name does not match provided package name'
                    raise ValueError(message)
                if 'version' not in meta:
                    message = 'meta file missing version value'
                    raise ValueError(message)
                return meta
        except JSONDecodeError:
            message = 'invalud package file format'
            raise ValueError(message)

    @staticmethod
    def _remove_files(path: Path) -> None:
        """Recursively remove the path and all its sub items."""
        path = str(path)[7:] if str(path)[:7] == 'file://' else path
        try:
            shutil.rmtree(path)
        except OSError:
            FileNotFoundError('unable to remove files')


class HttpPackageHandler(PackageHandler):
    """Url package structure."""
    def __init__(
        self,
        name: str,
        src: str,
        formats: Formats = None,
        cache_dir: str = None
    ):
        """initialise the package."""
        self._meta = {}
        self._meta['name'] = name
        self._path = None
        self._installed = False
        self._src = src
        if not cache_dir:
            tmp_cache = tempfile.TemporaryDirectory()
            cache_dir = Path(tmp_cache.name)
            atexit.register(tmp_cache.cleanup)
        else:
            cache_dir.mkdir(0x777, True, True)
        parts = urlparse(src)
        ext = ''.join(Path(parts.path).suffixes)
        self._cache = Path(cache_dir / Path(name + ext))
        if not HttpPackageHandler.grasps(src):
            message = 'invalid handler format'
            raise ValueError(message)
        if formats is None:
            formats = [ZipPackageFormat]
        for current_format in formats:
            if current_format.grasps(self._src):
                self._format = current_format
                break
        else:
            message = 'a valid source is required'
            raise FileNotFoundError(message)

    @property
    def name(self) -> str:
        """Return the name of the package from the meta data."""
        return self.meta['name']

    def __getattr__(self, name):
        """Return an attribute from the meta data if the data doesnt exist."""
        if name not in self._meta:
            message = f'{name} is an invalid attribute'
            raise AttributeError(message)
        return self._meta[name]

    @property
    def installed(self) -> bool:
        """Return the current installation state."""
        return self._installed

    @staticmethod
    def grasps(path: any) -> bool:
        """Check if a directory on the local filesystem has been provided."""
        parts = urlparse(str(path))
        if parts.scheme not in ['http', 'https']:
            return False
        return True

    def install(self, path: str, version: str = None) -> bool:
        """The local package handler does not install the package. No version
        control exists for the directory package type.
        """
        self._path = path
        self._updated = False
        headers = {'If-None-Match': self._digest}
        req = requests.get(self._src, headers=headers, allow_redirects=True)
        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            self._cache.parents[0].mkdir(parents=True, exist_ok=True)
            with open(self._cache, 'wb') as cached_package:
                cached_package.write(req.content)
            self._updated = True
        self._format.extract(self._cache, self._path)
        self._meta = {**self._meta, **self._load_metadata(self._path)}
        self._installed = True
        return True

    def uninstall(self) -> bool:
        """Uninstall the package."""
        if self._path != self._src:
            self._remove_files(self._path)
        self._path = None
        self._installed = False
        return True

    def verify(self, digest: str = None) -> bool:
        """Verify the package hasn't been tampered with."""
        if digest is None:
            return True
        message = 'the http package handler does not support verification'
        raise ValueError(message)

    def _load_metadata(self, path: Path):
        """Load the package.json file into memory."""
        try:
            with open(path / 'package.json') as package:
                meta = json.load(package)
                if 'name' not in meta:
                    message = 'meta file missing name value'
                    raise ValueError(message)
                if self._meta['name'] != meta['name']:
                    message = 'meta name does not match provided package name'
                    raise ValueError(message)
                if 'version' not in meta:
                    message = 'meta file missing version value'
                    raise ValueError(message)
                return meta
        except JSONDecodeError:
            message = 'invalud package file format'
            raise ValueError(message)

    @staticmethod
    def _remove_files(path: Path) -> None:
        """Recursively remove the path and all its sub items."""
        try:
            shutil.rmtree(path)
        except OSError:
            FileNotFoundError('unable to remove files')

    @property
    def outdated(self) -> bool:
        """Execute a head request using the requests library to check that the
        ETag matches.
        """
        headers = {'If-None-Match': self._digest}
        req = requests.head(self._src, headers=headers, allow_redirects=True)
        if req.status_code not in [200, 304]:
            raise FileNotFoundError(req.status_code)
        elif req.status_code == 200:
            return True
        elif req.status_code == 304:
            return False

    @property
    def _digest(self) -> str:
        """Return the md5 digest of the currently cached index file."""
        digest = md5()  # noqa: S303
        if not self._cache.exists():
            return digest.hexdigest()
        with open(self._cache, "rb") as cached_package:
            for block in iter(lambda: cached_package.read(65536), b""):
                digest.update(block)
        return digest.hexdigest()



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import logging
from logging import NullHandler
from . import _plugins


log = logging.getLogger(__name__).addHandler(NullHandler())


class plugin:  # noqa: N801
    """Plugin decorator to register plugins."""

    def __init__(self, name, priority=100):
        """Get the name of the registered method to add the plugin to and set
        the evecution priority. Priority less than zero executes before the
        main method, priority above zero executes afterwards.
        """
        log.debug('plugin.__init__: ({}) {}'.format(priority, name))

        # ensure the priority is numeric
        if not isinstance(priority, int):
            raise ValueError('Priority needs to be numeric')
        self._priority = priority

        # check the name is printable
        if not name.isprintable():
            raise ValueError('Name needs to be printable')

        # check the plugin registration exists
        self._name = name
        if name not in _plugins:
            log.warning('function "{}" not registered'.format(name))
            _plugins[name] = {'original': None}
            _plugins[name]['plugins'] = []

    def __call__(self, extend_function):
        """"""
        log.debug('plugin.__call__: {}'.format(extend_function))
        self._function = extend_function
        global _plugins
        _plugins[self._name]['plugins'].append({
            'function': self._function,
            'priority': self._priority
        })



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dismantle',
 'dismantle.extension',
 'dismantle.index',
 'dismantle.package',
 'dismantle.plugin']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.1,<3.0.0']

entry_points = \
{'console_scripts': ['area28 = dismantle.cli:main']}

setup_kwargs = {
    'name': 'dismantle',
    'version': '0.7.0',
    'description': 'Python package / plugin / extension manager',
    'long_description': '# Dismantle\n\n**Dismantle** Python package / plugin / extension manager.\n\n```python\n"""Simple plugin example"""\nfrom dismantle import plugin\n\n\nclass Chat():\n    @plugin.register(\'chat.message\')\n    def show(self, message):\n        print(message)\n\n\n@plugin.plugin(\'chat.message\', order=-1)\ndef make_uppercase(message):\n    return message.upper()\n\n```\n\n```python\n"""Full example using all aspects."""\nfrom pathlib import Path\nfrom dismantle.extension import Extensions, IExtension\nfrom dismantle.index import JsonFileIndexHandler\nfrom dismantle.package import LocalPackageHandler\n\n\nclass ColorExtension(IExtension):\n    _category = \'color\'\n\n    def color(self) -> None:\n        ...\n\n\nclass GreenColorExtension(ColorExtension):\n    _name = \'green\'\n\n    def color(self) -> None:\n        print(f\'color is {self._name}\')\n\n\npackages = {}\nindex = JsonFileIndexHandler(\'index.json\')\nfor pkg_meta in index:\n    meta = index[pkg_meta]\n    path = datadir.join(meta[\'path\'])\n    package = LocalPackageHandler(meta[\'name\'], path)\n    package._meta = {**package._meta, **meta}\n    package.install()\n    packages[package.name] = package\nextensions = Extensions([ColorExtension], packages, \'ext_\')\nassert extensions.types == [\'color\']\nassert list(extensions.category(\'color\').keys()) == [\n    \'@scope-one/package-one.extension.green.GreenColorExtension\',\n    \'@scope-one/package-two.extension.red.RedColorExtension\',\n    \'@scope-one/package-three.extension.blue.BlueColorExtension\',\n]\nassert list(extensions.extensions.keys()) == [\'color\']\nassert list(extensions.imports.keys()) == [\n    \'@scope-one/package-one.extension.green\',\n    \'@scope-one/package-two.extension.red\',\n    \'@scope-one/package-three.extension.blue\'\n]\n```\n\nDismantle allows you to provide the ability to create a plugin/extension/module for an application.\nIt does this by checking a package index and using that index to manage package versions. Packages\nthen contain plugins (using decorators) and extensions (using a custom module loader) to add the\nadditional functionality to the application.\n\n## Installing Dismantle and Supported Versions\n\nDismantle is available on PyPI:\n\n```console\n$ python -m pip install dismantle\n```\n\nDismantle officially supports Python 3.7+.\n\n## Supported Features & Best–Practices\n\nDismantle is ready for the demands of providing flexibility within applications allowing developers\nto build rich ecosystems around core applications.\n\n### Index Management\n\n- easy to create custom index handlers providing additional ways to define package indexes\n- local index file support using json\n- url based index file support using json\n- etag based caching for url based index\n\n### Packaging\n\n- easy to create custom package handlers providing additional ways to define package sources\n- easy to create custom package formats compression types and structures\n- support for zip, tar.gz, tgz, and local directories as package formats\n- support for local and url based (http/https) package handlers\n- hash validation for packages with the ability to verify package integrity\n\n### Extensions\n\n- Categorized extension groups to filter extension types (eg. loggers, parsers, ...)\n- Support for __init__ or .py based module loading.\n- Extension activation and deactivation management.\n- Module name collision avoidance\n- Hierarchical module naming\n\n### Plugins\n\n- Decorator based plugins with pre and post value modification\n- Multiple plugins per function with ability to set execution order\n- Activation management\n',
    'author': 'Gary Stidston-Broadbent',
    'author_email': 'dismantle@garysb.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/garysb/dismantle',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Copyright 2021 Gary Stidston-Broadbent
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import functools
import logging
from logging import NullHandler
from . import _plugins


log = logging.getLogger(__name__).addHandler(NullHandler())


def register(name=None):
    """Register a function to add plugins to."""
    log.debug('register: {}'.format(name))
    if not name:
        raise ValueError

    def decorator(_func=None):
        log.debug('register.decorator: {}'.format(_func))
        _plugins[name] = {'original': _func}
        _plugins[name]['plugins'] = []

        @functools.wraps(_func)
        def caller(self, *args, **kwargs):
            """ Execute the registered plugin in order of priority. """

            # store the previous result
            _plugins[name]['result'] = args[0]

            # get the list of before and after plugins
            before = [p for p in _plugins[name]['plugins'] if p['priority'] < 0]  # NOQA: E501
            after = [p for p in _plugins[name]['plugins'] if p['priority'] >= 0]  # NOQA: E501

            # execute all the before plugins
            for _name in before:
                _args = (self, _plugins[name]['result'])
                _plugins[name]['result'] = _name['function'](*_args)

            # execute the actual method that was registered
            _args = (self, _plugins[name]['result'])
            _plugins[name]['result'] = _plugins[name]['original'](*_args)

            # execute all the after plugins
            for _name in after:
                _args = (self, _plugins[name]['result'])
                _plugins[name]['result'] = _name['function'](*_args)

            return _plugins[name]['result']
        return caller
    return decorator
